package multex; //MsgText.java



//Change history:
//2001-02-26  Knabe  Internationalization of message texts is now optional.
//2000-10-25  Knabe  appendMessageChain and appendMessageLine append to a
//                   StringBuffer rather than returning a String.
//2000-09-29  Knabe  Unnamed parameters are no longer substituted into a
//                   message text, only named ones (public nostatic fields)
//2000-07-24  Knabe  Umsetzung von Namensparametern mittels Reflection
//2000-07-07  Knabe  extrahiert aus Klasse Msg


/** Services for getting the message text for exceptions.
  It handles exceptions <UL>
    <LI>with parameters (subclasses of {@link Exc})
    <LI>with cause chain and parameters (subclasses of {@link Failure})
    <LI>and standard Java exceptions (other subclasses of java.lang.Throwable)
  </UL>
  @author Christoph Knabe, Berlin, Copyright 1999-2000
*/
public class MsgText {


private static final String _className = new MsgText().getClass().getName();
private static final String _resourceName = "MsgText";
private static final String _because = "CAUSE: ";

/**Activates the internationalization of message texts. Further calls to
  {@link #appendMessageLine} or {@link #appendMessageChain} will deliver the
  internationalized message text for an exception instead of the default message
  text contained in the MultexException-object.
  @since MulTEx 1c
*/
public static void setInternationalization(final boolean i_internationalization){
    _internationalization = i_internationalization;
    _messages = _internationalization ? _tryLoadMessages() : null;
}

/**Appends the message for i_throwable with its causal chain.
  Appends an usually multi-line message, in which the first line is the
    message for i_throwable and the further lines are the messages in the
    chain of causing exceptions of i_throwable.
    See {@link #appendMessageLine} for a despription,
    how each exception is transformed into a message line.
  @param i_lineSeparator The String used to separate the lines of the message chain.
    If null is provided, defaults to {@link Failure#lineSeparator}.
*/
public static void appendMessageChain(
  final StringBuffer io_buffer,
  final Throwable i_throwable,
  final String i_lineSeparator
){
  //Throwable-Meldungen eintragen:
  for(Throwable t=i_throwable; t!=null; t=Failure.getCause(t)){
    if(t!=i_throwable){ //causing exception
      io_buffer.append(
          i_lineSeparator!=null ? i_lineSeparator : Failure.lineSeparator
      );
      io_buffer.append(_because);
    }
    appendMessageLine(io_buffer, t);
  }
}


/**Appends the usually one-line message text for the exception
  i_throwable with inserted parameters to io_buffer.

  The message text pattern for the exception is taken from the mostly localized
  variant of the properties file "MsgText.properties" or helpwise from the
  default message text pattern in the exception object itself. A local message
  text pattern for the german language (de) in its swiss (CH) variant would be
  taken from file "MsgText_de_CH.properties", see the Java Tutorial
  <A HREF="http://java.sun.com/docs/books/tutorial/i18n/format/messageFormat.html">
  chapter on internationalization</A> at february 1999.
  <P>
  This file must contain definition lines of the form: <PRE>
     packageName.ExceptionName = message text pattern
  </PRE>or<PRE>
     packageName.ClassName$InnerExceptionName = message text pattern
  </PRE>with parameter substitutors in the message text pattern, e.g. with named parameters:<PRE>
     CopyNP$CopyFailure = The file "{0}" could not be copied to "{1}"
  </PRE>
  ???If i_throwable is a standard Java exception with one String as
  detailMessage, this is taken as message parameter {0}.
*/
public static void appendMessageLine(final StringBuffer io_buffer, final Throwable i_throwable){
  final int startOfThisLine = io_buffer.length();
  try {
    final String textPattern = _getMessageTextPattern(i_throwable);
    final Object[] parameters = _getParameters(i_throwable);
    final String formattedMessage;
    if(parameters==null){ //no insertable exception parameters
      formattedMessage = textPattern;
    }else{ //with exception parameters:
      formattedMessage = java.text.MessageFormat.format(
        textPattern, parameters
      );
    }
    io_buffer.insert(startOfThisLine, formattedMessage);
    if(!(i_throwable instanceof MultexException)){
      //other Java exceptions may have detailMessage:
      final String detailMessage = i_throwable.getMessage();
      if(detailMessage!=null && detailMessage.length()!=0){
        //Append standard Java detailMessage, separated by a colon
        io_buffer.append(": ");
        io_buffer.append(detailMessage);
      }
    }
  } catch(final Exception ex){
      System.err.println();
      System.err.println(_className + ": EXCEPTION OCCURRED WHEN PREPARING MESSAGE LINE");
      ex.printStackTrace();
      io_buffer.append(Failure.lineSeparator);
      io_buffer.append(i_throwable.toString());
    }
  //end try
}//appendMessageLine

/**Gets a message text pattern for the exception i_throwable.
  The following sources are tried one after another with the class name
  of i_throwable as message text key: <OL>
    <LI>The language and country specific MsgText properties file,
        e.g. MsgText_en_UK.properties
    <LI>The language specific MsgText properties file,
        e.g. MsgText_en.properties
    <LI>The default MsgText properties file: MsgText.properties
    <LI>The default message text pattern from i_throwable, if it is a
        MultexException
    <LI>The class name of the exception i_throwable
  </OL>
*/
private static String _getMessageTextPattern(final Throwable i_throwable){
  final String className = i_throwable.getClass().getName();
  final String pattern0 = i_throwable instanceof MultexException ?
    ((MultexException)i_throwable).getDefaultMessageTextPattern() : className
  ;
  final String defaultPattern = pattern0==null ? className : pattern0;
  //2001-02-26  Knabe  Will now be done in setInternationalization(true);  :
  //_tryLoadMessages();
  if(_messages==null){
    return defaultPattern;
  }
  try{return _messages.getString(className);}
    catch(java.util.MissingResourceException ex){
      return defaultPattern;
    }
  //try
}//_getMessageTextPattern

/**Returns the parameters contained in i_throwable or null*/
private static Object[] _getParameters(final Throwable i_throwable){
  if(!(i_throwable instanceof MultexException)){return null;}
  return ((MultexException)i_throwable).getParameters();
}

/*debug
private static void _println(final String i_name, final Object[] i_array){
  System.err.print(i_name + ": ");
  for(int i=0; i<i_array.length; i++){
    System.err.print(i_array[i] + "  ");
  }
  System.err.println();
}
*/


/**The maximum number of index parameters, as allowed by
  java.text.MessageFormat
*/
private static final int _maxNumberOfIndexParams = 10;


//Message text access:

private static boolean                  _internationalization = false;
private static java.util.ResourceBundle _messages;
private static boolean                  _loadMessagesTried = false;

/**If _internalionalization is on, tries once to load the ResourceBundle
  _resourceName as message text file and returns it.
  If the loading fails, this method will return null,
  and will do nothing on further invocations.
*/
private static synchronized java.util.ResourceBundle _tryLoadMessages(){
  //if(!_internationalization){return;}
  //if(_loadMessagesTried){return;}
  try{
    _loadMessagesTried = true;
    return java.util.ResourceBundle.getBundle(_resourceName);
  } catch(final java.util.MissingResourceException ex){ //resource not found
      System.err.println(ex);
      System.err.println(_className + ": Unable to load exception message texts"
        + " from the ResourceBundle\n" + _resourceName
      );
      return null;
    } catch(final Exception ex){ //not loadable
      ex.printStackTrace();
      return null;
    }
  //end try
}//_tryLoadMessages


/**Do not create objects of me!*/
private MsgText(){}

}//MsgText
